;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "hl-todo" "3.9.3"
  "Highlight TODO and similar keywords."
  '((emacs  "26.1")
    (compat "30.1"))
  :url "https://github.com/tarsius/hl-todo"
  :commit "9540fc414014822dde00f0188b74e17ac99e916d"
  :revdesc "v3.9.3-0-g9540fc414014"
  :keywords '("convenience")
  :authors '(("Jonas Bernoulli" . "emacs.hl-todo@jonas.bernoulli.dev"))
  :maintainers '(("Jonas Bernoulli" . "emacs.hl-todo@jonas.bernoulli.dev")))
